@@ -17,8 +17,6 @@
 
 (define applicable? arity)
 
-;(define sequence? length) -- now built-in for signatures
-
 (define (nth-value n . args) ; untested...
   (and (> (length args) n)
        (list-ref args n)))
@@ -48,9 +46,9 @@
 	     (if (not (null? info))
 		 (format p ": ~A" info))))
 
-       (format p "~%error-code: ~S~%" (ow 'error-code))
+       (format p "~%error-code: ~S" (ow 'error-code))
        (when (ow 'error-line)
-	 (format p "~%error-file/line: ~S[~A]~%" (ow 'error-file) (ow 'error-line)))
+	 (format p "~%error-file/line: ~A[~A]" (ow 'error-file) (ow 'error-line)))
 	   
        ;; show history, if available
        (when (pair? (ow 'error-history)) ; a circular list, starts at error-code, entries stored backwards
@@ -108,15 +106,20 @@
 	       vals)
      ,@(map (lambda (val)
 	      (if (pair? (cddr val))
-		  `(set! (setter ',(car val)) 
-			 (lambda (s v)
-			   (if (not (,(caddr val) v))
-			       (error 'wrong-type-arg "(set! ~S ~S) but ~S is not ~A" s v v ',(caddr val)))
-			   v))
+		  `(set! (setter ',(car val))
+			 (if (not (,(caddr val) ,(car val))) ; check initial value (already set)
+			     (error 'wrong-type "initial value ~S is not ~S" ,(car val) ,(caddr val))
+			     ,(caddr val)))                  ; assume built-in type here
 		  (values)))
 	    vals)
      ,@body))
 
+(define-macro (typed-inlet . vals) ; vals: ((var init [type])...)...) as in (typed-inlet (i 0 integer?)...)
+  (let ((e (gensym)))
+    `(let ((,e (typed-let ,vals (curlet))))
+       (set! (outlet ,e) (rootlet)) ; mimic (inlet ...)
+       ,e)))
+
 
 ;;; ----------------
 (define (first obj)  (if (sequence? obj) (obj 0) (error 'wrong-type-arg "first argument, ~S, is not a sequence" obj)))
@@ -133,10 +136,22 @@
 (define (built-in? x) 
   (not (undefined? (eval-string (string-append "#_" (object->string x)))))) ; just a guess...
 
+#|
 (define (the type expr) 
   (if (type expr)
       expr
       (error 'bad-type "~S is ~S but should be ~S" expr (type-of expr) type)))
+|#
+(define (the type expr)
+  (let ((e expr)
+	(bp type))
+    (if (and (procedure? bp)
+	     (signature bp)
+	     (eq? 'boolean? (car (signature bp))))
+	(if (type e)
+	    e
+	    (error 'bad-type "~S is ~S but should be ~S" e (type-of e) bp))
+	(error 'bad-type "~S is not a boolean procedure" bp))))
 
 (define iota 
   (let ((+documentation+ "(iota n (start 0) (incr 1)) returns a list counting from start for n:\n\
@@ -333,7 +348,9 @@
 |#
 
 (define-macro (and-let* vars . body)      ; bind vars, if any is #f stop, else evaluate body with those bindings
-  `(let () (and ,@(map (lambda (v) (cons 'define v)) vars) (begin ,@body))))
+  (if (list? vars)
+      `(let () (and ,@(map (lambda (v) (cons 'define v)) vars) (begin ,@body)))
+      (error 'wrong-type-arg "and-let* var list is ~S" vars)))
 
 (define-macro (let*-temporarily vars . body)
   `(with-let (#_inlet :orig (#_curlet) 
@@ -354,8 +371,6 @@
 			  `(set! (with-let orig ,(car v)) (list-ref saved ,(set! ctr (+ ctr 1)))))))
 		  vars)))))
 
-;;; 14-8-16: moved let-temporarily to s7.c
-
 (define-macro (let/setter vars . body)
   (let ((setters (map (lambda (binding)
 			 (and (pair? (cddr binding))
@@ -428,13 +443,7 @@
 	  (map values table)
 	  (error 'wrong-type-arg "hash-table->alist argument, ~A, is not a hash-table" table)))))
 
-(define merge-hash-tables 
-  (let ((+documentation+ "(merge-hash-tables . tables) returns a new hash-table with the contents of all the tables"))
-    (lambda tables
-      (apply hash-table 
-	     (apply append 
-		    (map hash-table->alist tables))))))
-
+(define merge-hash-tables append)
 
 
 ;;; ----------------
@@ -528,7 +537,9 @@ If func approves of one, index-if returns the index that gives that element's po
     (lambda (type f sequence)
       (unless (sequence? sequence)
 	(error 'wrong-type-arg "collect-if: sequence arg is ~A" sequence))
-      (apply type (map (lambda (arg) (if (f arg) arg (values))) sequence)))))
+      (if (eq? type hash-table)
+	  (apply hash-table (map (lambda (arg) (if (f arg) (values (car arg) (cdr arg)) (values))) sequence))
+	  (apply type (map (lambda (arg) (if (f arg) arg (values))) sequence))))))
 
 ;;; if type=list, this is slightly wasteful because list currently copies its args, so:
 ;;;   ((if (eq? type list) values (values apply type)) ...) would work
@@ -699,7 +710,7 @@ Unlike full-find-if, safe-find-if can handle any circularity in the sequences.")
 ;;; ----------------
 (define sequences->list 
   (let ((+documentation+ "(sequences->list . sequences) returns a list of elements of all the sequences:\n\
-    (sequences->list \"hi\" #(0 1) (hash-table '(a . 2))) -> '(#\\h #\\i 0 1 (a . 2))"))
+    (sequences->list \"hi\" #(0 1) (hash-table 'a 2)) -> '(#\\h #\\i 0 1 (a . 2))"))
     (lambda sequences
       (apply append 
 	     (map (lambda (sequence) 
@@ -737,7 +748,9 @@ Unlike full-find-if, safe-find-if can handle any circularity in the sequences.")
 				(if (not (member obj lst))
 				    (set! lst (cons obj lst))))
 			      (apply sequences->list sequences))
-		    (reverse lst))))))
+		    (if (eq? type hash-table)
+			(map (lambda (x) (values (car x) (cdr x))) (reverse lst))
+			(reverse lst)))))))
 
 (define asymmetric-difference 
   (let ((+documentation+ "(asymmetric-difference type . sequences) returns the elements in the rest of the sequences that are not in the first:\n\
@@ -914,17 +927,13 @@ Unlike full-find-if, safe-find-if can handle any circularity in the sequences.")
 (define (byte siz pos) ;; -> cache size, position and mask.
   (list siz pos (ash (- (ash 1 siz) 1) pos)))
 
-(define byte-size car)
-(define byte-position cadr)
-(define byte-mask caddr)
-
 (define (ldb bytespec integer)
-  (ash (logand integer (byte-mask bytespec))
-       (- (byte-position bytespec))))
+  (ash (logand integer (caddr bytespec))
+       (- (cadr bytespec))))
 
 (define (dpb integer bytespec into)
-  (logior (ash (logand integer (- (ash 1 (byte-size bytespec)) 1)) (byte-position bytespec))
-	  (logand into (lognot (byte-mask bytespec)))))
+  (logior (ash (logand integer (- (ash 1 (car bytespec)) 1)) (cadr bytespec))
+	  (logand into (lognot (caddr bytespec)))))
 
 
 ;;; ----------------
@@ -976,8 +985,9 @@ Unlike full-find-if, safe-find-if can handle any circularity in the sequences.")
 		   
 		   ;; add an object->string method for this class (this is already a generic function).
 		   (list (cons 'object->string 
-			       (lambda* (obj (use-write #t))
-				 (if (eq? use-write :readable)    ; write readably
+			       (lambda (obj . rest)
+				 (if (and (pair? rest)
+					  (eq? (car rest) :readable))    ; write readably
 				     (format #f "(make-~A~{ :~A ~W~^~})" 
 					     ',class-name 
 					     (map (lambda (slot)
@@ -1174,6 +1184,8 @@ Unlike full-find-if, safe-find-if can handle any circularity in the sequences.")
       (if (continuation? ((owlet) 'continue))
 	  (((owlet) 'continue))))))
 
+;;; ----------------
+
 (define (call-with-input-vector v proc)
   (if (vector? v)
       (let ((i -1))
@@ -1201,7 +1213,7 @@ Unlike full-find-if, safe-find-if can handle any circularity in the sequences.")
 			    (if (not (let? p))
 				(apply format p args)
 				(write (apply format #f args) p))))))
-    (subvector v (list i)))) ; ignore extra trailing elements
+    (subvector v i))) ; ignore extra trailing elements (i = subvector length)
 
 
 
@@ -1351,7 +1363,7 @@ Unlike full-find-if, safe-find-if can handle any circularity in the sequences.")
 	(error 'out-of-range "end: ~A should be greater than start: ~A" end start))
 
     (cond ((vector? obj) 
-	   (subvector obj (list new-len) start))
+	   (subvector obj new-len start))
 
           ((string? obj)
            (if (integer? end)
@@ -1372,19 +1384,6 @@ Unlike full-find-if, safe-find-if can handle any circularity in the sequences.")
 	       ((= i new-len) lst)
 	     (set! (lst i) (obj (+ i start))))))))
 
-
-(define (sequence->string val)
-  (let ((ctrl-str (cond ((or (not (sequence? val)) 
-			     (empty? val))    "~S")
-			((vector? val)        "#(~{~A~| ~})")
-			((let? val)           "(inlet ~{'~A~| ~})")
-			((hash-table? val)    "(hash-table ~{'~A~| ~})")
-			((byte-vector? val)   "#u8(~{~D~| ~})")
-			((not (string? val))  "(~{~A~| ~})")
-			(else                 "\"~{~A~|~}\""))))
-    (format #f ctrl-str val)))
-
-
 ;;; ----------------
 
 (define-macro (elambda args . body)  ; lambda but pass extra arg "*env*" = run-time env
@@ -1402,350 +1401,33 @@ Unlike full-find-if, safe-find-if can handle any circularity in the sequences.")
 
 ;;; ----------------
 
-(if (not (or (defined? 'apropos)
-	     (provided? 'snd)))
-    (define* (apropos name (port *stdout*) (e (rootlet)))
-      (let ((ap-name (if (string? name) 
-			 name 
-			 (if (symbol? name) 
-			     (symbol->string name)
-			     (error 'wrong-type-arg "apropos argument 1 should be a string or a symbol"))))
-	    (ap-env (if (let? e) 
-			e 
-			(error 'wrong-type-arg "apropos argument 3 should be an environment")))
-	    (ap-port (if (output-port? port) 
-			 port
-			 (error 'wrong-type-arg "apropos argument 2 should be an output port"))))
-	(for-each
-	 (lambda (binding)
-	   (if (and (pair? binding)
-		    (string-position ap-name (symbol->string (car binding))))
-	       (format ap-port "~A: ~A~%" 
-		       (car binding) 
-		       (if (procedure? (cdr binding))
-			   (documentation (cdr binding))
-			   (cdr binding)))))
-	 ap-env))))
-
-
-;;; ----------------
-
-;; these need to be globally accessible since they're inserted in arbitrary source
-(define Display #f)
-(define Display-port #f)
-
-(let ((spaces 0)
-      (*display-spacing* 2)         ; (set! ((funclet Display) '*display-spacing*) 1) etc
-      (*display-print-length* 6)
-      (*display* *stderr*)          ; exported via Display-port
-      (e (gensym))
-      (result (gensym))
-      (vlp (gensym)))
-  
-  ;; local symbol access -- this does not affect any other uses of these symbols
-  (set! (setter '*display-spacing* (curlet))
-	(lambda (s v) (if (and (integer? v) (not (negative? v))) v *display-spacing*)))
-  
-  (set! (setter '*display-print-length* (curlet))
-	(lambda (s v) (if (and (integer? v) (not (negative? v))) v *display-print-length*)))
-  
-  ;; export *display* -- just a convenience
-  (set! Display-port (dilambda
-		      (lambda () *display*)
-		      (lambda (val) (set! *display* val))))
-  
-  (define (prepend-spaces)
-    (format *display* (format #f "~~~DC" spaces) #\space))
-  
-  (define (display-let le e)
-    (let ((vlp (*s7* 'print-length)))
+(unless (or (defined? 'apropos)
+	    (provided? 'snd))
+  (define* (apropos name (port *stdout*) (e (rootlet)))
+    (let ((ap-name (if (string? name) 
+		       name 
+		       (if (symbol? name) 
+			   (symbol->string name)
+			   (error 'wrong-type-arg "apropos argument 1 should be a string or a symbol"))))
+	  (ap-env (if (let? e) 
+		      e 
+		      (error 'wrong-type-arg "apropos argument 3 should be an environment")))
+	  (ap-port (if (output-port? port) 
+		       port
+		       (error 'wrong-type-arg "apropos argument 2 should be an output port"))))
       (for-each
-       (lambda (slot)
-	 (unless (or (gensym? (car slot))
-		     (eq? (cdr slot) e))
-	   (set! (*s7* 'print-length) *display-print-length*)
-	   (let ((str (sequence->string (cdr slot))))
-	     (set! (*s7* 'print-length) 30)
-	     (format (Display-port) " :~A ~{~A~|~}" (car slot) str))))
-       le)
-      (set! (*s7* 'print-length) vlp)))
-  
-  (define (last lst)
-    (let ((end (let ((len (length lst)))
-		 (list-tail lst (if (negative? len) (abs len) (- len 1))))))
-      (if (pair? end)
-	  (car end)
-	  end)))
-  
-  (define* (butlast lst (result ()))
-    (if (or (not (pair? lst))
-	    (null? (cdr lst)))
-	(reverse result)
-	(butlast (cdr lst) (cons (car lst) result))))
-  
-  (define* (remove-keys args (lst ()))
-    (if (pair? args)
-	(remove-keys (cdr args) 
-		     (if (or (not (keyword? (car args)))
-			     (eq? (car args) :rest))
-			 (cons (car args) lst)
-			 lst))
-	(if (null? args)
-	    (reverse lst)
-	    (append (reverse lst) args))))
-  
-  (define (proc-walk source)
-    
-    (define (display-format str . args)
-      `(let ((,vlp (*s7* 'print-length)))
-	 (with-let (funclet Display)
-	   (set! (*s7* 'print-length) *display-print-length*)
-	   (prepend-spaces))
-	 (format (Display-port) ,str ,@args)
-	 (set! (*s7* 'print-length) ,vlp)))
-  
-    (define (walk-let-body source)
-      (let ((previous (butlast source))
-	    (end (last source)))
-	`(begin
-	   ,@previous
-	   (let ((,result ,end))
-	     (with-let (funclet Display)
-	       (prepend-spaces))
-	     (format (Display-port) "  ~A~A) -> ~A~%"
-		     ,(if (pair? previous) " ... " "")
-		     ',end
-		     ,result)
-	     ,result))))
-    
-    (if (not (pair? source))
-	source
-	(case (car source)
-	  
-	  ((let let* letrec letrec*)                
-	   ;; show local variables, (let vars . body) -> (let vars print-vars . body)
-	   (if (symbol? (cadr source))           ; named let?
-	       (append 
-		(list (car source)               ; let
-		      (cadr source)              ; name
-		      (caddr source)             ; vars
-		      (display-format "(let ~A (~{~A~| ~})~%" (cadr source) '(outlet (curlet))))
-		(walk-let-body (cdddr source)))      ; body
-	       (append 
-		(list (car source)               ; let
-		      (cadr source)              ; vars
-		      (display-format "(~A (~{~A~| ~})~%" (car source) '(outlet (curlet))))
-		(walk-let-body (cddr source)))))     ; body
-	  
-	  ((or and)
-	   ;; report form that short-circuits the evaluation
-	   (cons (car source)
-		 (let ((ctr -1)
-		       (len (- (length source) 2))
-		       (eob (if (eq? (car source) 'or) 'when 'unless)))
-		   (map (lambda (expr)
-			  (set! ctr (+ ctr 1))
-			  `(let ((,result ,expr))
-			     (,eob ,result
-				   (format (Display-port) "  (~A ~A~A~A) -> ~A~%" 
-					   ',(car source)
-					   ,(if (> ctr 0) " ... " "")
-					   ',expr 
-					   ,(if (< ctr len) " ... " "")
-					   ,result))
-			     ,result))
-			(cdr source)))))
-	  
-	  ((begin with-let with-baffle)
-	   ;; report last form
-	   (let ((previous (butlast (cdr source)))
-		 (end (last source)))
-	     `(,(car source)
-	       ,@previous
-	       (let ((,result ,end))
-		 (format (Display-port) "(~A ~A~A) -> ~A~%"
-			 ',(car source)
-			 ,(if (pair? previous) " ... " "")
-			 ',end
-			 ,result)
-		 ,result))))
-	  
-	  ((when unless)
-	   ;; report expr if body not walked, else report last form of body
-	   (let ((previous (butlast (cddr source)))
-		 (end (last (cddr source))))
-	     `(,(car source) (let ((,result ,(cadr source)))
-			       (,(car source) (not ,result)
-				(format (Display-port) "(~A ~A -> ~A ...)~%"
-					',(car source)
-					',(cadr source)
-					,result))
-			       ,result)
-	       ,@previous
-	       (let ((,result ,end))
-		 (format (Display-port) "(~A ... ~A) -> ~A~%"
-			 ',(car source)
-			 ',end
-			 ,result)
-		 ,result))))
-	  
-	  ((quote)
-	   source)
-	  
-	  ((cond)
-	   ;; report form that satisifies cond
-	   (let ((ctr -1)
-		 (len (- (length source) 2)))
-	     `(cond ,@(map (lambda (clause)
-			     (let ((test (car clause))
-				   (body (cdr clause)))
-			       (set! ctr (+ ctr 1))
-			       (if (eq? (car body) '=>) 
-				   `(,test => (lambda (,result) 
-						(let ((,result (,@(cdr body) ,result)))
-						  (format (Display-port) "  (cond ~A~A~A) -> ~A~%" 
-							  ,(if (> ctr 0) " ... " "")
-							  ',clause
-							  ,(if (< ctr len) " ... " "")
-							  ,result)
-						  ,result)))
-				   `(,test (let ((,result (begin ,@body)))
-					     (format (Display-port) "  (cond ~A~A~A) -> ~A~%" 
-						     ,(if (> ctr 0) " ... " "")
-						     ',clause
-						     ,(if (< ctr len) " ... " "")
-						     ,result)
-					     ,result)))))
-			   (cdr source)))))
-	  
-	  ((case)
-	   ;; as in cond but include selector value in [] and report fall throughs
-	   (let ((ctr -1)
-		 (len (- (length source) 3))
-		 (default (member '(else #t) (cddr source) (lambda (a b) 
-							     (memq (car b) a)))))
-	     `(case ,(cadr source)
-		,@(append 
-		   (map (lambda (clause)
-			  (let ((test (car clause))
-				(body (cdr clause)))
-			    (set! ctr (+ ctr 1))
-			    (if (eq? (car body) '=>)
-				`(,test => (lambda (,result)
-					     (let ((,result (,@(cdr body) ,result)))
-					       (format (Display-port) "  (case [~A] ~A~A~A) -> ~A~%" 
-						       ,(cadr source)
-						       ,(if (> ctr 0) " ... " "")
-						       ',clause
-						       ,(if (< ctr len) " ... " "")
-						       ,result)
-					       ,result)))
-				`(,test (let ((,result (begin ,@body)))
-					  (format (Display-port) "  (case [~A] ~A~A~A) -> ~A~%" 
-						  ,(cadr source)
-						  ,(if (> ctr 0) " ... " "")
-						  ',clause
-						  ,(if (< ctr len) " ... " "")
-						  ,result)
-					  ,result)))))
-			(cddr source))
-		   (if default
-		       ()
-		       `((else 
-			  (format (Display-port) "  (case [~A] falls through~%" ,(cadr source)) 
-			  #<unspecified>)))))))
-	  
-	  ((dynamic-wind)
-	   ;; here we want to ignore the first and last clauses, and report the last of the second
-	   (let ((body (let ((p (caddr source)))
-			 (and (eq? (car p) 'lambda)
-			      (cddr p)))))
-	     (if (not body)
-		 source
-		 (let ((previous (and body (butlast body)))
-		       (end (and body (last body))))
-		   `(dynamic-wind
-			,(cadr source)
-			(lambda ()
-			  ,@previous
-			  (let ((,result ,end))
-			    (format (Display-port) "(dynamic-wind ... ~A) -> ~A~%" ',end ,result)
-			    ,result))
-			,(cadddr source))))))
-	  
-	  (else
-	   (cons (proc-walk (car source)) 
-		 (proc-walk (cdr source)))))))
-
-  
-  (define-macro (Display-1 definition)
-    (if (not (and (pair? definition)
-		  (memq (car definition) '(define define*))
-		  (pair? (cdr definition))
-		  (pair? (cadr definition))))
-
-	;; (Display <anything-else>)
-	(proc-walk definition)                                                    ; (Display (+ x 1)) etc
-	
-	;; (Display (define (f ...) ...)
-	(let ((func (caadr definition))
-	      (args (cdadr definition))
-	      (body (cons 'begin (proc-walk (cddr definition)))))
-	  ;(format *stderr* "~A ~A ~A~%" func args body)
-	  (let* ((no-noise-args (remove-keys args))                               ; omit noise words like :optional
-		 (arg-names (cond ((null? args)
-				   ())
-				  ((proper-list? args)                            ; handle (f x ...), (f (x 1) ...), (f . x), and (f x . z)
-				   (map (lambda (a) 
-					  (if (symbol? a) a (car a)))             ; omit the default values
-					no-noise-args))                                
-				  ((pair? args)
-				   (append (butlast no-noise-args) (list :rest (last args))))
-				  (else (list :rest args))))
-		 (call-args (cond ((null? args)
-				   ())
-				  ((proper-list? args)
-				   (if (memq :rest args)
-				       (append (butlast (butlast no-noise-args))   ; also omit the :rest
-					       (list (list 'apply-values (last args))))
-				       arg-names))                                 ; (... y x)
-				  ((pair? args)
-				   (append (butlast no-noise-args)            ; (... y (apply-values x))
-					   (list (list 'apply-values (last args)))))
-				  (else (list (list 'apply-values args))))))     ; (... (apply-values x))
-	    `(define ,func
-	       (define-macro* ,(cons (gensym) args)                                ; args might be a symbol etc
-		 `((lambda* ,(cons ',e ',arg-names)                                ; prepend added env arg because there might be a rest arg
-		     (let ((,',result '?))
-		       (dynamic-wind
-			   (lambda () #f)
-			   (lambda ()                                              ; when function called, show args and caller
-			     (with-let (funclet Display)                           ; indent
-			       (prepend-spaces)
-			       (set! spaces (+ spaces *display-spacing*)))
-			     (format (Display-port) "(~A" ',',func)                ; show args, ruthlessly abbreviated
-			     (((funclet Display) 'display-let) (outlet (outlet (curlet))) ,',e)
-			     (format (Display-port) ")")
-			     (let ((caller (eval '__func__ ,',e)))                 ; show caller 
-			       (if (not (eq? caller #<undefined>))
-				   (format (Display-port) " ;called from ~A" caller)))
-			     (newline (Display-port))
-			     (set! ,',result ,',body))                             ;   but annotated by proc-walk
-			   (lambda ()                                              ; at the end, show the result
-			     (with-let (funclet Display)
-			       (set! spaces (- spaces *display-spacing*))  ; unindent
-			       (prepend-spaces))
-			     (format (Display-port) "    -> ~S~%" ,',result)))))
-		   (curlet) ,,@call-args)))))))                                    ; pass in the original args and the curlet
-	
-  
-  (set! Display Display-1))                                                        ; make Display-1 globally accessible
-
-
+       (lambda (binding)
+	 (if (and (pair? binding)
+		  (string-position ap-name (symbol->string (car binding))))
+	     (format ap-port "~A: ~A~%" 
+		     (car binding) 
+		     (if (procedure? (cdr binding))
+			 (documentation (cdr binding))
+			 (cdr binding)))))
+       ap-env))))
 
 ;;; --------------------------------------------------------------------------------
 
-
 (define (*s7*->list) 
   (list 
    :print-length                  (*s7* 'print-length)
@@ -1765,7 +1447,7 @@ Unlike full-find-if, safe-find-if can handle any circularity in the sequences.")
    :initial-string-port-length    (*s7* 'initial-string-port-length)
    :default-rationalize-error     (*s7* 'default-rationalize-error)
    :default-random-state          (*s7* 'default-random-state)
-   :morally-equal-float-epsilon   (*s7* 'morally-equal-float-epsilon)
+   :equivalent-float-epsilon   (*s7* 'equivalent-float-epsilon)
    :hash-table-float-epsilon      (*s7* 'hash-table-float-epsilon)
    :float-format-precision        (*s7* 'float-format-precision)
    :bignum-precision              (*s7* 'bignum-precision)
@@ -1778,7 +1460,11 @@ Unlike full-find-if, safe-find-if can handle any circularity in the sequences.")
    :max-stack-size                (*s7* 'max-stack-size)
    :autoloading?                  (*s7* 'autoloading?)
    :undefined-identifier-warnings (*s7* 'undefined-identifier-warnings)
-   :catches                       (*s7* 'catches)))
+   :undefined-constant-warnings   (*s7* 'undefined-constant-warnings)
+   :catches                       (*s7* 'catches)
+   :history-size                  (*s7* 'history-size)
+   :history-enabled               (*s7* 'history-enabled)
+   :history                       (*s7* 'history)))
 
 
 
@@ -1865,7 +1551,7 @@ Unlike full-find-if, safe-find-if can handle any circularity in the sequences.")
 	   (for-each
 	    (lambda (op) 
 	      (set! (ht op) #t))
-	    '(symbol? gensym? keyword? let? openlet? iterator? macro? c-pointer? c-object? immutable? constant?
+	    '(symbol? gensym? keyword? let? openlet? iterator? macro? c-pointer? c-object? c-object-type immutable? constant?
 	      input-port? output-port? eof-object? integer? number? real? complex? rational? random-state? 
 	      char? string? list? pair? vector? float-vector? int-vector? byte-vector? hash-table? 
 	      continuation? procedure? dilambda? boolean? float? proper-list? sequence? null? gensym 
@@ -1896,11 +1582,11 @@ Unlike full-find-if, safe-find-if can handle any circularity in the sequences.")
 	      list->vector vector-fill! vector-length vector->list vector-ref vector-set! vector-dimensions 
 	      make-vector subvector vector float-vector make-float-vector float-vector-set! 
 	      float-vector-ref int-vector make-int-vector int-vector-set! int-vector-ref string->byte-vector 
-	      byte-vector make-byte-vector hash-table hash-table* make-hash-table hash-table-ref 
+	      byte-vector make-byte-vector hash-table make-hash-table hash-table-ref weak-hash-table
 	      hash-table-set! hash-table-entries cyclic-sequences call/cc call-with-current-continuation 
 	      call-with-exit apply for-each map dynamic-wind values type-of
 	      catch throw error documentation signature help procedure-source
-	      setter arity aritable? not eq? eqv? equal? morally-equal? s7-version
+	      setter arity aritable? not eq? eqv? equal? equivalent? s7-version
 	      dilambda make-hook hook-functions stacktrace tree-leaves tree-memq tree-set-memq tree-cyclic? tree-count object->let
 	      pi most-positive-fixnum most-negative-fixnum nan.0 inf.0 -nan.0 -inf.0
 	      *stderr* *stdout* *stdin*
@@ -1925,6 +1611,13 @@ Unlike full-find-if, safe-find-if can handle any circularity in the sequences.")
 			 #_getenv #_system #_delete-file #_directory->list #_directory? #_file-exists? #_file-mtime))
 		       )))
     (lambda (code)
+      (if (tree-cyclic? code)
+	  (error 'wrong-type-arg "sandbox argument is circular: ~S~%" code))
+      (if (and (pair? code)
+	       (>= (length code) 10000))
+	  (error 'wrong-type-arg
+		 (let-temporarily (((*s7* 'print-length) 16))
+		   (format #f "sandbox code looks bogus: ~S~%" code))))
       ;; block any change to calling program, or access to files, etc
       (let ((new-code 
 	     (call-with-exit
@@ -1955,6 +1648,11 @@ Unlike full-find-if, safe-find-if can handle any circularity in the sequences.")
 			((not (pair? tree))
 			 tree)
 			
+			((eq? 'quote (car tree)) ; tree-cyclic? ignores quoted lists
+			 (if (tree-cyclic? (cdr tree))
+			     (error 'wrong-type-arg "sandbox argument is circular: ~S~%" tree))
+			 tree)
+
 			(else
 			 ;; do we need to check IO ports and set! here?
 			 (cons (walk (car tree))
